"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""
import os
from pwd import getpwnam

from wsgidav import util
from wsgidav.middleware import BaseMiddleware

from fileman.middleware.impersonate.exception import ImpersonateException


_logger = util.getModuleLogger(__name__)


class ImpersonateFilter(BaseMiddleware):
    """
    Impersonates the current process down to the given user (with his euid and egid) so the lower layer operation will
    be done from the name of that user. After the next middleware finishes, this middleware takes back the root
    privileges.
    """

    def __init__(self, application, config):
        self._application = application
        self._config = config
        _logger.debug('impersonate filter created')

    def __call__(self, environ, start_response):
        _logger.debug("Processing request from PID %s" % (os.getpid()))
        sub_app_start_response = util.SubAppStartResponse()

        #
        # XXX [KN] IMPORTANT: forking server is assumed!
        #
        self.set_umask(environ)
        self.drop_privileges(environ)
        try:
            response_started = False
            app_iter = self._application(environ, sub_app_start_response)
            for v in app_iter:
                # Start response (the first time)
                if not response_started:
                    # Success!
                    start_response(sub_app_start_response.status,
                                   sub_app_start_response.response_headers,
                                   sub_app_start_response.exc_info)
                response_started = True

                yield v
        finally:
            self.gain_privileges()
            self.reset_umask(environ)

        # Close out iterator
        if hasattr(app_iter, "close"):
            app_iter.close()

        # Start response (if it hasn't been done yet)
        if not response_started:
            # Success!
            start_response(sub_app_start_response.status,
                           sub_app_start_response.response_headers,
                           sub_app_start_response.exc_info)

        return

    def gain_privileges(self):
        _logger.debug("Current privileges are %s:%s" % (os.geteuid(), os.getegid()))
        _logger.debug("Trying to get the root privileges back")
        try:
            os.seteuid(0)
            os.setegid(0)
        except OSError as e:
            _logger.error("Failed to become root")
            _logger.exception(e)
            raise
        _logger.debug("Root privileges are acquired successfully")

    def drop_privileges(self, environ):
        assert ('http_authenticator.username' in environ) and (environ["http_authenticator.username"] != '')
        username = environ["http_authenticator.username"]

        try:
            pw_entry = getpwnam(username)
        except KeyError:
            err_text = 'Invalid username passed for impersonation: %s' % username
            _logger.debug(err_text)
            raise ImpersonateException(err_text)

        uid = pw_entry.pw_uid
        gid = pw_entry.pw_gid

        _logger.debug("Current privileges are %s:%s" % (os.geteuid(), os.getegid()))
        _logger.debug("Dropping down privileges to %s:%s" % (uid, gid))

        try:
            os.setegid(gid)
            os.seteuid(uid)
        except OSError:
            _logger.error("Failed to drop privileges though the values are valid. Do you have correct privileges?")
            raise

    def set_umask(self, environ):
        new_umask = self.get_config_umask(environ)
        old_umask = os.umask(new_umask)

        _logger.debug("Changed umask from %s to %s" % (old_umask, new_umask))
        environ["default_umask"] = old_umask

    def reset_umask(self, environ):
        assert "default_umask" in environ
        old_umask = os.umask(environ["default_umask"])
        _logger.debug("Reverted umask to %s from %s" % (environ["default_umask"], old_umask))

    def get_config_umask(self, environ):
        config = environ["wsgidav.config"]
        return config.get('default_umask', 0000)
