"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""
import os
import sys
import zipfile

from wsgidav import util
from wsgidav.dav_error import DAVError, HTTP_BAD_REQUEST
from wsgidav.dav_provider import DAVNonCollection
from wsgidav.fs_dav_provider import FolderResource, FileResource

from fileman.middleware.filtering_filter import FILTERING_KEY, FILTER_BY_PARAM, FILTER_VALUE
from fileman.middleware.paging_filter import PAGING_KEY
from fileman.middleware.sorting_filter import ORDERING_KEY, ORDER_BY_PARAM, ORDER_DIRECTION_PARAM
from fileman.middleware.roots.parser import PathSplitter
from fileman.provider.impersonated.decorators import fs_permission_required
from fileman.provider.impersonated.directory_lister import DirectoryLister
from fileman.provider.impersonated.file_filter import ExactNamePredicate, ContainsNamePredicate, Filter
from fileman.provider.impersonated.sorter import NameSorter, CreateDateSorter, ExtensionSorter, SizeSorter
from fileman.provider.impersonated.mixins import ArchiveSupportMixin, AccessibleRootsHelper, SymlinkSupportMixin, \
    AcDavMixin
from fileman.provider.impersonated.utils import get_bool, to_bool, is_property_equal, walk_directory, enumerate_files, \
    get_home_dir, is_home_request, get_mime_type


OWNERGROUP = "{DAV:}ownergroup"
OWNERUSER = "{DAV:}owneruser"
PERMISSIONMASK = "{DAV:}permissionmask"
IS_ARCHIVE = "{DAV:}isarchive"
IS_SYMLINK = "{DAV:}issymlink"
IS_SYMLINK_OK = "{DAV:}issymlinkok"
ITEMSCOUNT = "{DAV:}itemscount"
ENCODING = "{DAV:}encoding"

_logger = util.getModuleLogger(__name__)


class ImpersonatedFolderResource(ArchiveSupportMixin, SymlinkSupportMixin, FolderResource):
    def getPropertyNames(self, isAllProp):
        props = super(ImpersonatedFolderResource, self).getPropertyNames(isAllProp)
        return props + [PERMISSIONMASK, OWNERUSER, OWNERGROUP, IS_ARCHIVE, IS_SYMLINK, IS_SYMLINK_OK, ITEMSCOUNT]

    @fs_permission_required
    def getPropertyValue(self, propname):
        if is_property_equal(propname, PERMISSIONMASK):
            return self.get_permission_mask()
        elif is_property_equal(propname, OWNERUSER):
            return self.get_owner_uid()
        elif is_property_equal(propname, OWNERGROUP):
            return self.get_owner_gid()
        elif is_property_equal(propname, IS_ARCHIVE):
            return get_bool(self._is_archive())
        elif is_property_equal(propname, IS_SYMLINK):
            return get_bool(self._is_symlink(self._filePath))
        elif is_property_equal(propname, IS_SYMLINK_OK):
            return get_bool(self._is_symlink_sane(self._filePath))
        elif is_property_equal(propname, ITEMSCOUNT):
            return self.count_child_items()

        return super(ImpersonatedFolderResource, self).getPropertyValue(propname)

    @fs_permission_required
    def setPropertyValue(self, propname, value, dryRun=False):
        if is_property_equal(propname, PERMISSIONMASK):
            mask = value.text
            if dryRun:
                return
            return self.set_permission_mask(mask)
        if is_property_equal(propname, IS_ARCHIVE):
            do_archive = to_bool(value.text)
            if dryRun:
                return self._validate_archive_prop_change(do_archive)
            return self._handle_archive_prop_change(do_archive)

        return super(ImpersonatedFolderResource, self).setPropertyValue(propname, value, dryRun)

    @fs_permission_required
    def createCollection(self, name):
        folder = super(ImpersonatedFolderResource, self).createCollection(name)
        return folder

    def getContent(self):
        pass

    def supportRanges(self):
        return False

    @fs_permission_required
    def moveRecursive(self, destPath):
        return super(ImpersonatedFolderResource, self).moveRecursive(destPath)

    @fs_permission_required
    def delete(self):
        return super(ImpersonatedFolderResource, self).delete()

    @fs_permission_required
    def createEmptyResource(self, name):
        return super(ImpersonatedFolderResource, self).createEmptyResource(name)

    @fs_permission_required
    def copyMoveSingle(self, destPath, isMove):
        return super(ImpersonatedFolderResource, self).copyMoveSingle(destPath, isMove)

    def _get_filter_instance(self):
        filtering_data = self.environ.get(FILTERING_KEY, None)
        filter_inst = None
        if filtering_data:
            filter_inst = self.create_filter(filtering_data)
        return filter_inst

    def get_directory_lister(self):
        paging_data = self.environ.get(PAGING_KEY, None)
        sorting_data = self.environ.get(ORDERING_KEY, None)
        sorter = None

        if sorting_data:
            sorter = self.create_sorter(sorting_data)

        filter_inst = self._get_filter_instance()

        lister = DirectoryLister(self._filePath, paging_data, self.environ, sorter=sorter, filter=filter_inst)
        return lister

    def list_directory_contents(self):
        lister = self.get_directory_lister()
        return lister.list_contents()

    @fs_permission_required
    def getMemberNames(self):
        name_list = []
        # self._filePath is unicode, so os.listdir returns unicode as well
        assert isinstance(self._filePath, unicode)
        for name in self.list_directory_contents():
            if not isinstance(name, unicode):
                name = name.decode(sys.getfilesystemencoding())
            assert isinstance(name, unicode)
            # Skip non files (links and mount points)
            fp = os.path.join(self._filePath, name)

            if not os.path.isdir(fp) and not os.path.isfile(fp) and not self._is_symlink(fp):
                _logger.debug("Skipping non-file %s" % fp)
                continue

            name = name.encode("utf8")
            name_list.append(name)
        return name_list


    # overriding just to ensure that correct classes for FileResource and FolderResources are used
    def create_folder_resource(self, fp, path):
        return ImpersonatedFolderResource(path, self.environ, fp)

    def create_file_resource(self, fp, path):
        return ImpersonatedFileResource(path, self.environ, fp)

    def handle_symlink(self, fp, path):
        if not self._is_symlink_sane(fp):
            res = self.create_file_resource(fp, path)
        else:
            target_path = self.readlink(fp)
            if os.path.isdir(target_path):
                res = self.create_folder_resource(fp, path)
            else:
                res = self.create_file_resource(fp, path)
        return res

    def getMember(self, name):
        """Return direct collection member (DAVResource or derived).

        See DAVCollection.getMember()
        """
        fp = os.path.join(self._filePath, name.decode("utf8"))
        # name = name.encode("utf8")
        path = util.joinUri(self.path, name)

        if os.path.isdir(fp):
            res = self.create_folder_resource(fp, path)

        elif os.path.isfile(fp):
            res = self.create_file_resource(fp, path)

        elif self._is_symlink(fp):
            res = self.handle_symlink(fp, path)

        else:
            _logger.debug("Skipping non-file %s" % fp)
            res = None
        return res

    def getPreferredPath(self):
        if is_home_request(self.path):
            username = self.environ["http_authenticator.username"]
            return get_home_dir(username)
        return super(ImpersonatedFolderResource, self).getPreferredPath()

    def count_child_items(self):
        filter_inst = self._get_filter_instance()
        # [KN] Intentionally pass paging_data = None to ensure that we'll count
        # all the element (no paging is applied).
        lister = DirectoryLister(self._filePath, None, self.environ, filter=filter_inst)
        value = sum(1 for x in lister.list_contents())
        return str(value)

    def create_sorter(self, sorting_data):
        order_by = sorting_data[ORDER_BY_PARAM]
        direction = sorting_data[ORDER_DIRECTION_PARAM]
        is_ascending = direction == 'ASC'

        if order_by == 'getlastmodified':
            _logger.debug('Sorting by modification date will be performed')
            return CreateDateSorter(is_asc=is_ascending)

        if order_by == 'displayname':
            _logger.debug('Sorting by name will be performed')
            return NameSorter(is_asc=is_ascending)

        if order_by == 'extension':
            _logger.debug('Sorting by extension will be performed')
            return ExtensionSorter(is_asc=is_ascending)

        if order_by == 'getcontentlength':
            _logger.debug('Sorting by size will be performed')
            return SizeSorter(is_asc=is_ascending)

        return None

    def create_filter(self, filtering_data):
        filter_by = filtering_data[FILTER_BY_PARAM]
        filter_value = filtering_data[FILTER_VALUE]

        predicate_map = {'exact_name': ExactNamePredicate, 'contains_name': ContainsNamePredicate}

        pred = predicate_map.get(filter_by, None)
        if not pred:
            return None

        return Filter(pred(filter_value))


def patch_file_resource():
    """
    Fixes (patches) the __init__ method in FileResource class (see wsgidav library sources) to ensure that broken
    symlinks are processed correctly.
    :return:
    """
    global FileResource

    def __init__(self, path, environ, filePath):
        super(FileResource, self).__init__(path, environ)
        self._filePath = filePath

        # XXX this is what the patch is created for
        self.filestat = self.get_stat()

        # Setting the name from the file path should fix the case on Windows
        self.name = os.path.basename(self._filePath)
        self.name = self.name.encode("utf8")

    class_dict = dict(FileResource.__dict__)
    FileResource = type('FileResource', (DAVNonCollection,), class_dict)
    FileResource.__init__ = __init__


patch_file_resource()


class ImpersonatedFileResource(FileResource, ArchiveSupportMixin, SymlinkSupportMixin, AcDavMixin):
    def __init__(self, path, environ, filePath):
        super(ImpersonatedFileResource, self).__init__(path, environ, filePath)

    def getPropertyNames(self, isAllProp):
        props = super(ImpersonatedFileResource, self).getPropertyNames(isAllProp)
        return props + [PERMISSIONMASK, OWNERUSER, OWNERGROUP, IS_ARCHIVE, IS_SYMLINK, IS_SYMLINK_OK, ENCODING]

    def getContentType(self):
        return get_mime_type(self.path)


    @fs_permission_required
    def getContent(self):
        return super(ImpersonatedFileResource, self).getContent()

    @fs_permission_required
    def getPropertyValue(self, propname):
        if is_property_equal(propname, PERMISSIONMASK):
            return self.get_permission_mask()
        elif is_property_equal(propname, OWNERUSER):
            return self.get_owner_uid()
        elif is_property_equal(propname, OWNERGROUP):
            return self.get_owner_gid()
        elif is_property_equal(propname, IS_ARCHIVE):
            return get_bool(self._is_archive())
        elif is_property_equal(propname, IS_SYMLINK):
            return get_bool(self._is_symlink(self._filePath))
        elif is_property_equal(propname, IS_SYMLINK_OK):
            return get_bool(self._is_symlink_sane(self._filePath))
        elif is_property_equal(propname, ENCODING):
            return self.get_file_encoding()

        return super(ImpersonatedFileResource, self).getPropertyValue(propname)

    def get_file_encoding(self):
        return utils.get_file_encoding(self._filePath)

    @fs_permission_required
    def setPropertyValue(self, propname, value, dryRun=False):
        if is_property_equal(propname, PERMISSIONMASK):
            mask = value.text
            if dryRun:
                return
            return self.set_permission_mask(mask)

        if is_property_equal(propname, IS_ARCHIVE):
            do_archive = to_bool(value.text)
            if dryRun:
                return self._validate_archive_prop_change(do_archive)
            return self._handle_archive_prop_change(do_archive)

        return super(ImpersonatedFileResource, self).setPropertyValue(propname, value, dryRun)

    @fs_permission_required
    def beginWrite(self, contentType=None):
        return super(ImpersonatedFileResource, self).beginWrite(contentType)

    @fs_permission_required
    def endWrite(self, withErrors):
        return super(ImpersonatedFileResource, self).endWrite(withErrors)

    @fs_permission_required
    def moveRecursive(self, destPath):
        return super(ImpersonatedFileResource, self).moveRecursive(destPath)

    @fs_permission_required
    def createEmptyResource(self, name):
        return super(ImpersonatedFileResource, self).createEmptyResource(name)

    @fs_permission_required
    def delete(self):
        return super(ImpersonatedFileResource, self).delete()

    @fs_permission_required
    def copyMoveSingle(self, destPath, isMove):
        return super(ImpersonatedFileResource, self).copyMoveSingle(destPath, isMove)

    def create_group_archive(self, archive_path, path_list):
        dest_name = os.path.normpath(archive_path)
        dir_name = os.path.dirname(archive_path)
        rel_paths = map(lambda v: self.to_relative_path(v, dir_name), path_list)
        old_dir = os.getcwd()
        os.chdir(dir_name)
        zipf = zipfile.ZipFile(dest_name, 'w', compression=zipfile.ZIP_DEFLATED)

        for next_file in enumerate_files(rel_paths):
            absolute_path = os.path.join(dir_name, next_file)
            is_empty_dir = os.path.isdir(absolute_path)
            if is_empty_dir:
                """
                Given the zipfile module in python it's not obvious how to add an empty directory to a zip file.
                Adding an empty directory to a zip file requires that you add a trailing slash to an empty file or
                set the external_attribute for directories (16).
                """
                zfi = zipfile.ZipInfo(next_file + '/')
                zipf.writestr(zfi, '')
            else:
                _logger.debug('Adding file to archive: %s' % next_file)
                zipf.write(next_file)

        zipf.close()

        os.chdir(old_dir)
        _logger.debug('Group archive operation is finished')

    def to_relative_path(self, abs_path, new_root):
        splitter = PathSplitter()
        abs_fragments = splitter.parse(abs_path)
        root_fragments = splitter.parse(new_root)

        for fragment in root_fragments:
            if abs_fragments[0] != fragment:
                raise DAVError(HTTP_BAD_REQUEST)
            abs_fragments.pop(0)
        return '/'.join(abs_fragments)


