"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""
import os
import shutil
import tempfile
import zipfile
import re

from wsgidav import util
from wsgidav.dav_error import HTTP_BAD_REQUEST, DAVError, HTTP_NOT_ACCEPTABLE
from wsgidav.dav_provider import DAVProvider, DAVNonCollection, DAVCollection

import lxml.etree as etree
from fileman.middleware.roots.parser import PathSplitter
from fileman.provider.impersonated import fs_permission_required, AcDavMixin, enumerate_files
from fileman.provider.impersonated.utils import rstrip_url, to_relative_path


_logger = util.getModuleLogger(__name__)


class BaseBulkOperation(AcDavMixin, DAVNonCollection):
    """
    Base class for all bulk file operations. The subclasses have to implement _perform_bulk_operation()
    and _get_filename_key() methods.
    """
    def delete(self):
        pass

    def _get_filename_key(self):
        """
        :return: A const string value to be used as a key in self.environ dictionary. This is where the temporary file
        name will be stored until the whole upload finishes.
        """
        raise NotImplementedError()

    def _perform_bulk_operation(self, path_list, target_path):
        """
        The central method where the actual logic of bulk operation is contained.
        :param path_list: The list of file paths to operate over.
        :param target_path: The target file name (not always applicable).
        :return:
        """
        raise NotImplementedError()

    @fs_permission_required
    def beginWrite(self, contentType=None):
        fobject = tempfile.NamedTemporaryFile(delete=False)
        self.environ[self._get_filename_key()] = fobject.name
        return fobject

    def extract_path_list(self, body_xml):
        elem_list = body_xml.xpath("/bulk/objects/object/@path")
        target_path = body_xml.xpath("/bulk/target/@path")

        target_path = target_path[0] if len(target_path) else None
        path_list = map(self._remove_share_prefix, elem_list)
        return path_list, target_path

    @fs_permission_required
    def endWrite(self, withErrors):
        _logger.debug('The data necessary for bulk operation is acquired.')

        file_name = self.environ[self._get_filename_key()]
        try:
            body_xml = etree.parse(file_name)
            path_list, target_path = self.extract_path_list(body_xml)
            self._perform_bulk_operation(path_list, target_path)
            return

        finally:
            if os.path.exists(file_name):
                os.unlink(file_name)


class StubCollection(DAVCollection):
    """
    Stub collection object. Used to mimic as if the Bulk*Operation objects have a parent.
    """
    pass


class BulkDeleteOperation(BaseBulkOperation):
    def _get_filename_key(self):
        return 'bulk.delete.filename'

    def _perform_bulk_operation(self, path_list, target_path):
        _logger.debug('The following files will be removed: %s' % path_list)
        self._remove_files(path_list)

    def _remove_files(self, path_list):
        for file_item in path_list:
            file_item = self._sanitize_path(file_item)
            if os.path.isdir(file_item):
                shutil.rmtree(file_item, ignore_errors=False)
            else:
                os.unlink(file_item)

    def _sanitize_path(self, file_path):
        regexp = re.compile(r'/+$')
        if regexp.search(file_path):
            # delete all slashes at the end of the url
            file_path = regexp.sub('', file_path)
        return file_path


class BulkArchiveOperation(BaseBulkOperation):
    def _get_filename_key(self):
        return 'bulk.archive.filename'

    def _perform_bulk_operation(self, path_list, target_path):
        self.create_group_archive(target_path, path_list)

    def create_group_archive(self, archive_path, path_list):
        dest_name = os.path.normpath(archive_path)
        dir_name = os.path.dirname(archive_path)
        rel_paths = map(lambda v: self.to_relative_path(v, dir_name), path_list)
        old_dir = os.getcwd()
        os.chdir(dir_name)
        zipf = zipfile.ZipFile(dest_name, 'w', compression=zipfile.ZIP_DEFLATED)

        for next_file in enumerate_files(rel_paths):
            absolute_path = os.path.join(dir_name, next_file)
            is_empty_dir = os.path.isdir(absolute_path)
            if is_empty_dir:
                """
                Given the zipfile module in python it's not obvious how to add an empty directory to a zip file.
                Adding an empty directory to a zip file requires that you add a trailing slash to an empty file or
                set the external_attribute for directories (16).
                """
                zfi = zipfile.ZipInfo(next_file + '/')
                zipf.writestr(zfi, '')
            else:
                _logger.debug('Adding file to archive: %s' % next_file)
                zipf.write(next_file)

        zipf.close()

        os.chdir(old_dir)
        _logger.debug('Group archive operation is finished')

    def to_relative_path(self, abs_path, new_root):
        try:
            return to_relative_path(abs_path, new_root)
        except ValueError:
            raise DAVError(HTTP_BAD_REQUEST)


class BulkCopyOperation(BaseBulkOperation):
    def _get_filename_key(self):
        return 'bulk.copy.filename'

    def _perform_bulk_operation(self, path_list, target_path):
        _logger.debug('Bulk copy operation is started')
        target_path = os.path.normpath(target_path)
        if not os.path.isdir(target_path):
            raise DAVError(HTTP_NOT_ACCEPTABLE)

        for next_file in path_list:
            next_file = os.path.normpath(next_file)
            _logger.debug("Copying %s to %s" % (next_file, target_path))
            if os.path.isdir(next_file):
                _logger.debug("Notice: %s is a folder. Copying it recursively" % next_file)
                src_dir = os.path.basename(next_file)
                new_target = os.path.join(target_path, src_dir)
                if os.path.exists(new_target): 
                    new_target = os.path.join(target_path, "Copy 1 of " + src_dir)
                    index = 1 
                    while os.path.exists(new_target):
                        index += 1                         
                        new_target = os.path.join(target_path, "Copy " + str(index) + " of " + src_dir) 
                    os.makedirs(new_target)
                    for item in os.listdir(next_file):
                        old_path = os.path.join(next_file, item)
                        new_path = os.path.join(new_target, item)
                        if os.path.isdir(old_path):
                            shutil.copytree(old_path, new_path, symlinks=True)
                        else:
                            shutil.copy2(old_path, new_path)                        
                else:
                    shutil.copytree(next_file, new_target, symlinks=True)               
            else:
                file_name = os.path.basename(next_file)
                new_target_path = os.path.join(target_path, file_name)
                if not os.path.exists(new_target_path):
                    shutil.copy2(next_file, target_path)
                else:                           
                    new_target_path = os.path.join(target_path, "Copy 1 of " + file_name)
                    index = 1   
                    while os.path.exists(new_target_path):
                        index += 1                         
                        new_target_path = os.path.join(target_path, "Copy " + str(index) + " of " + file_name)
                    shutil.copy2(next_file, new_target_path)
        _logger.debug('Bulk copy operation is finished')


class BulkMoveOperation(BaseBulkOperation):
    def _get_filename_key(self):
        return 'bulk.move.filename'

    def _perform_bulk_operation(self, path_list, target_path):
        _logger.debug('Bulk move operation is started')
        target_path = os.path.normpath(target_path)
        if not os.path.isdir(target_path):
            raise DAVError(HTTP_NOT_ACCEPTABLE)

        for next_file in path_list:
            _logger.debug("Moving %s to %s" % (next_file, target_path))
            shutil.move(next_file, target_path)

        _logger.debug('Bulk move operation is finished')


class BulkOperationProvider(DAVProvider):
    def getResourceInst(self, path, environ):
        stripped_path = rstrip_url(path)
        url_map = {'/files/delete': BulkDeleteOperation,
                   '/files/archive': BulkArchiveOperation,
                   '/files/copy': BulkCopyOperation,
                   '/files/move': BulkMoveOperation,
                   '/files': StubCollection}

        inst_class = url_map.get(stripped_path, None)
        if not inst_class:
            return None
        return inst_class(path, environ)
