"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""

from __future__ import absolute_import
import os

from wsgidav import util

from fileman.middleware.paging_filter import AMOUNT_PARAM, OFFSET_PARAM
from fileman.provider.impersonated.file_filter import Filter, TruePredicate
from fileman.provider.impersonated.mixins import AccessibleRootsSupportMixin, SymlinkSupportMixin, \
    HiddenFilesSupportMixin
from fileman.provider.impersonated.sorter import NameSorter, Sorter


_logger = util.getModuleLogger(__name__)


class DirectoryLister(AccessibleRootsSupportMixin, SymlinkSupportMixin, HiddenFilesSupportMixin, object):
    def __init__(self, path, paging_data, environ, sorter=None, filter=None):
        assert os.path.isdir(path)
        self.path = path
        self.paging_data = paging_data

        # [KN] environ is necessary for mixins to work.
        self.environ = environ
        self.sorter = sorter or NameSorter()
        self.filter = filter or self._create_noop_filter()

        assert isinstance(self.sorter, Sorter)
        assert isinstance(self.filter, Filter)

        super(DirectoryLister, self).__init__()

    def list_contents(self):
        dir_list = []
        file_list = []

        self._read_directory_contents(dir_list, file_list)
        # [KN] Pre-sorting is done to guarantee the stable order of items provided to consequent sorters.

        dir_list = list(self.filter.filter_out(dir_list))
        file_list = list(self.filter.filter_out(file_list))

        # Without it the sorters will rely on os.listdir() ordering which is arbitrary.
        dir_list = self.pre_sort(self.path, dir_list)
        file_list = self.pre_sort(self.path, file_list)

        dir_list = self.sorter.sort(self.path, dir_list)
        file_list = self.sorter.sort(self.path, file_list)

        if self.paging_data:
            amount = self.paging_data[AMOUNT_PARAM]
            offset = self.paging_data[OFFSET_PARAM]

            cur_idx = 0

            for d in dir_list:
                if cur_idx >= offset and amount > 0:
                    amount -= 1
                    yield d
                cur_idx += 1

            for f in file_list:
                if cur_idx >= offset and amount > 0:
                    amount -= 1
                    yield f
                cur_idx += 1
        else:
            for d in dir_list:
                yield d

            for f in file_list:
                yield f

    def _read_directory_contents(self, dir_list, file_list):
        elem_list = os.listdir(self.path)
        for el in elem_list:
            full_path = os.path.join(self.path, el)

            if not self.accessible_roots_satisfy(full_path):
                _logger.debug("Skipping folder due to accessible root setting %s" % full_path)
                continue

            elif not self.symlinks_policy_satisfy(full_path):
                _logger.debug("Skipping symlink file %s since symlinks are disabled by the configuration" % full_path)
                continue

            elif not self.hidden_objects_policy_satisfy(full_path):
                _logger.debug(
                    "Skipping hidden object %s since hidden objects are disabled by the configuration" % full_path)
                continue

            if os.path.isdir(full_path):
                dir_list.append(el)
            else:
                file_list.append(el)

    def pre_sort(self, path, dir_list):
        sorter = NameSorter()
        return sorter.sort(path, dir_list)

    def _create_noop_filter(self):
        filter = Filter(TruePredicate())
        return filter
