import json
import os
import re
import zipfile
import errno

from wsgidav.dav_error import DAVError, HTTP_PRECONDITION_FAILED

import magic
from fileman.middleware.roots.accessible_roots import ACCESSIBLE_ROOTS_KEY
from fileman.middleware.roots.matcher import Matcher
from fileman.provider.impersonated.utils import walk_directory, get_home_dir, _sanitize_unicode_path, is_archive

"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""


class OverwriteFailedError(RuntimeError):
    def __init__(self, *args, **kwargs):
        self._file_list = kwargs.pop('files', [])
        super(OverwriteFailedError, self).__init__(*args, **kwargs)

    def get_conflict_files(self):
        return self._file_list

    def __str__(self):
        return "CONFLICT_DATA=%s" % json.dumps(self._file_list)


class AcDavMixin(object):
    def _remove_share_prefix(self, value):
        value = re.sub(r'^/+(.*)$', r'/\1', value)
        share_path = self.provider.sharePath
        if share_path == '/' or not share_path:
            return value
        prefix = re.escape(share_path)
        regex = re.compile('^' + prefix + '(.*)$')
        path = re.sub(regex, r'/\1', value)
        return path


class ArchiveSupportMixin(object):
    def _should_archive(self, destPath):
        fname = os.path.basename(os.path.normpath(destPath))
        _, file_ext = os.path.splitext(fname)

        cur_path = os.path.basename(os.path.normpath(self._filePath))
        _, cur_ext = os.path.splitext(cur_path)

        return (file_ext == '.zip') and (cur_ext != file_ext)

    def _should_unarchive(self, destPath):
        fname = os.path.basename(os.path.normpath(destPath))
        _, file_ext = os.path.splitext(fname)

        cur_path = os.path.basename(os.path.normpath(self._filePath))
        _, cur_ext = os.path.splitext(cur_path)

        return (cur_ext == '.zip') and (file_ext != file_ext)

    def _unarchive(self, destPath, isMove):
        destPath = os.path.normpath(destPath)

        zipf = zipfile.ZipFile(self._filePath, 'r')

        # [KN] Note that zipf.extractall() is not used intentionally due to its problem with Unicode
        # Seems like when it is used, the directory is taken from os.getcwd() and it potentially might be non-unicode.
        # If the file name is unicode, the operation will fail with UnicodeDecodeError.

        for zipinfo in zipf.namelist():
            zipf.extract(zipinfo, destPath)

        if isMove:
            self.delete()

    def _is_archive(self):
        return is_archive(self._filePath)

    def generate_archive_name(self, fname):
        return fname + '.zip'

    def _handle_archive_prop_change(self, do_archive):
        if do_archive == self._is_archive():
            # nothing to do
            return
        fname = os.path.normpath(self._filePath)
        if do_archive:
            arch_name = self.generate_archive_name(fname)
            self._archive(arch_name, False)
            return
        else:
            dest_dir = os.path.dirname(fname)
            self._unarchive(dest_dir, False)

    def is_overwrite_enabled(self):
        return self.environ.get('HTTP_OVERWRITE', 'T') == 'T'

    def create_exists_exception(self, conflict_files):
        if not isinstance(conflict_files, list):
            conflict_files = [conflict_files]

        src_exception = OverwriteFailedError(files=conflict_files)
        return DAVError(HTTP_PRECONDITION_FAILED, srcexception=src_exception)

    def _validate_archive_prop_change(self, do_archive):
        if self.is_overwrite_enabled():
            return

        fname = os.path.normpath(self._filePath)
        if do_archive:
            archive_name = self.generate_archive_name(fname)
            if os.path.exists(archive_name):
                raise self.create_exists_exception(archive_name)
        else:
            dest_dir = os.path.dirname(fname)
            zipf = zipfile.ZipFile(fname, mode='r')
            conflict_files = []
            for item_name in zipf.namelist():
                if not os.path.isabs(item_name):
                    item_name = os.path.join(dest_dir, item_name)
                if os.path.exists(item_name):
                    conflict_files.append(item_name)

            if len(conflict_files):
                raise self.create_exists_exception(conflict_files)

    def get_file_list(self, fname):
        is_dir = os.path.isdir(fname)
        if not is_dir:
            return [fname]
        return walk_directory(fname)

    def _archive(self, destPath, isMove):
        destPath = os.path.normpath(destPath)
        dir = os.path.dirname(self._filePath)
        fname = os.path.basename(self._filePath)

        old_dir = os.getcwd()
        os.chdir(dir)
        zipf = zipfile.ZipFile(destPath, 'w', compression=zipfile.ZIP_DEFLATED)

        for file in self.get_file_list(fname):
            absolute_path = os.path.join(dir, file)
            is_empty_dir = os.path.isdir(absolute_path)
            if is_empty_dir:
                """
                Given the zipfile module in python it's not obvious how to add an empty directory to a zip file.
                Adding an empty directory to a zip file requires that you add a trailing slash to an empty file or
                set the external_attribute for directories (16).
                """
                zfi = zipfile.ZipInfo(file + '/')
                zipf.writestr(zfi, '')
            else:
                zipf.write(file)

        zipf.close()

        if isMove:
            self.delete()

        os.chdir(old_dir)


class SymlinkSupportMixin(object):
    def _is_symlink(self, file_path):
        path = os.path.normpath(file_path)
        result = os.path.islink(path)
        return result

    def readlink(self, file_path):
        """
        Returns the target absolute path where the given symlinks references to.
        :param file_path: Symlink location
        :return:
        """
        norm_path = os.path.normpath(file_path)
        target_path = os.readlink(norm_path)
        if not os.path.isabs(target_path):
            # Relative path is constructed starting from the current (self._filePath)

            parent_path = os.path.normpath(os.path.join(norm_path, '..'))
            target_path = os.path.join(parent_path, target_path)
            target_path = os.path.normpath(target_path)
        return target_path

    def _is_symlink_sane(self, file_path):
        """
        Checks whether the current File/Directory is a sane (non-broken) symlink.
        :return: False if the current object is not a symlink,
        False if the current object is a symlink which references to non-existing file object.
        True otherwise.
        """
        if not self._is_symlink(file_path):
            return False
        try:
            target_path = self.readlink(file_path)
        except OSError as e:
            if e.errno == errno.ELOOP:  # Looped symlink means broken
                return False
            raise

        return os.path.exists(target_path)

    def symlinks_policy_satisfy(self, file_path):
        if not self._is_symlink(file_path):
            return True
        return self.symlinks_enabled()

    def symlinks_enabled(self):
        """
        Checks whether the current configuration allows the symlinks.
        :return:
        """
        config = self.environ["wsgidav.config"]
        return config.get('allow_symlinks', False)

    def get_permission_mask(self):
        """
        :return: String representation of the given file permissions.
        """
        mask = self.get_stat()
        return oct(mask.st_mode & 0777)

    def get_stat(self):
        try:
            return os.stat(self._filePath)
        except OSError:
            # Do not follow symlink
            return os.lstat(self._filePath)

    def set_permission_mask(self, mask):
        """
        Sets the permission mask to the given file or directory.
        :param file_path: Path to the file or directory
        :param mask: The permission mask (lower 3 bits, e.g. 644) encoded as octal string (i.e. '0644').
        """
        int_mask = int(mask, base=8)
        curr_mask = self.get_stat().st_mode

        # ensure that the higher bits will remain the same.
        higher_bits = curr_mask & ~0777
        new_mask = higher_bits | int_mask

        os.chmod(self._filePath, new_mask)

    def get_owner_gid(self):
        return str(self.get_stat().st_gid)

    def get_owner_uid(self):
        return str(self.get_stat().st_uid)


class AccessibleRootsSupportMixin(object):
    matcher = Matcher()

    def get_matcher(self):
        return AccessibleRootsSupportMixin.matcher

    def get_accessible_roots(self):
        username = self.environ["http_authenticator.username"]
        home_dir = get_home_dir(username)
        return self.environ[ACCESSIBLE_ROOTS_KEY] + [home_dir]

    def accessible_roots_satisfy(self, path):
        if ACCESSIBLE_ROOTS_KEY not in self.environ:
            return True
        roots = self.get_accessible_roots()
        return self.get_matcher().match_roots(path, roots)


class HiddenFilesSupportMixin(object):
    def hidden_objects_policy_satisfy(self, fp):
        if not self.is_hidden(fp):
            return True
        return self.hidden_objects_enabled()

    def hidden_objects_enabled(self):
        config = self.environ["wsgidav.config"]
        return config.get('allow_hidden', False)

    def is_hidden(self, fp):
        file_name = os.path.basename(fp)
        return file_name.startswith('.') and (file_name not in ['.', '..'])


class AccessibleRootsHelper(AccessibleRootsSupportMixin):
    def __init__(self, environ):
        super(AccessibleRootsHelper, self).__init__()
        self.environ = environ
