"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""
from collections import deque
import os
import tempfile
import urlparse
from itertools import islice

from wsgidav import util

from wsgidav.dav_error import DAVError
from wsgidav.dav_provider import DAVNonCollection
from wsgidav.fs_dav_provider import FilesystemProvider

from fileman.provider.impersonated import AcDavMixin, AccessibleRootsHelper, fs_permission_required
from fileman.provider.impersonated.utils import get_mime_type


QUERY_STRING = 'QUERY_STRING'

_logger = util.getModuleLogger(__name__)


class TextPreviewOperation(AcDavMixin, DAVNonCollection):
    def __init__(self, path, environ, file_path):
        super(TextPreviewOperation, self).__init__(path, environ)
        self.file_path = file_path
        raw_value = environ[QUERY_STRING]
        q = urlparse.parse_qs(raw_value)

        self._assert(('direction' in q) and ('lines' in q))
        self._assert(os.path.isfile(self.file_path), error_code=404)

        direction = q['direction'][0]
        line_count = q['lines'][0]

        try:
            line_count = int(line_count)
        except ValueError:
            raise DAVError(400)

        self.temp_file_object = self.create_temp_file(self.file_path, direction, line_count)
        self.filePath = path

    def _assert(self, predicate, error_code=400):
        if not predicate:
            raise DAVError(error_code)

    def supportRanges(self):
        return False

    def getContent(self):
        return self.temp_file_object

    def getContentLength(self):
        file_object = self.temp_file_object
        return os.fstat(file_object.fileno()).st_size

    def getContentType(self):
        return get_mime_type(self.file_path)

    @fs_permission_required
    def create_temp_file(self, source_path, direction, line_count):
        _, tmp_name = tempfile.mkstemp(prefix='file-preview')
        f_object = open(tmp_name, mode='r+')
        fn = self._head
        if direction == 'tail':
            fn = self._tail

        with open(source_path, mode='r') as source_file:
            new_lines = fn(source_file, line_count)

        f_object.writelines(new_lines)
        f_object.seek(0)

        os.unlink(tmp_name)

        return f_object

    def _head(self, file_object, line_count=10):
        return list(islice(file_object, line_count))

    def _tail(self, file_object, line_count=10):
        lines_q = deque(maxlen=line_count)
        for line in file_object:
            lines_q.append(line)

        return ''.join(lines_q).strip()


class PreviewOperationProvider(FilesystemProvider):

    def isReadOnly(self):
        return True

    def getResourceInst(self, path, environ):
        self._count_getResourceInst += 1
        if isinstance(path, unicode):
            path = path.encode('utf-8')

        fp = self._locToFilePath(path)

        if not os.path.lexists(fp):
            _logger.debug("File doesn't exist: %s" % fp)
            return None

        # [KN] Since getResourceInst is called in from request_server.py:doGET method, we
        # should check whether the requested resource is available through the accessible roots.
        ar_helper = AccessibleRootsHelper(environ)
        if not ar_helper.accessible_roots_satisfy(fp):
            # In this case 404 Not Found is returned to the client.
            _logger.debug("Requested path is restricted by Accessible Roots policy: %s" % fp)
            return None

        if os.path.isdir(fp):
            return None
        return TextPreviewOperation(path, environ, fp)
