"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""
import os


class Sorter(object):
    def __init__(self, is_asc=True):
        self.is_asc = is_asc

    def sort(self, parent_folder, file_list):
        raise NotImplemented()


class NameSorter(Sorter):
    """
    Sorts the list of files or directories by their names (alphabetically).
    """

    def sort(self, parent_folder, file_list):
        desc = not self.is_asc
        file_list.sort(reverse=desc)
        return file_list


class StatBasedSorter(Sorter):
    def __init__(self, stat_fn=None, **kwargs):
        super(StatBasedSorter, self).__init__(**kwargs)
        self.get_stat = self._get_stat
        if stat_fn:
            self.get_stat = stat_fn

    def _get_stat(self, file_name):
        try:
            return os.stat(file_name)
        except OSError:
            # Do not follow symlink
            return os.lstat(file_name)


class CreateDateSorter(StatBasedSorter):

    def sort(self, parent_folder, file_list):
        stat = self.get_stat

        # [KN] stat() requires the full path to the file. file_list contains only short file names
        path = lambda x: os.path.join(parent_folder, x)

        # Create pairs (mtime, short_name)
        entries = map(lambda x: (stat(path(x)).st_ctime, x), file_list)

        # Sort by mtime
        # [KN] Note that the sort() is guaranteed to be stable (see
        # https://wiki.python.org/moin/HowTo/Sorting#Sort_Stability_and_Complex_Sorts)
        entries.sort(key=lambda x: x[0], reverse=not self.is_asc)

        # return the short names only
        return map(lambda x: x[1], entries)


class SizeSorter(StatBasedSorter):

    def sort(self, parent_folder, file_list):
        stat = self.get_stat

        # [KN] stat() requires the full path to the file. file_list contains only short file names
        path = lambda x: os.path.join(parent_folder, x)

        # Create pairs (st_size, short_name)
        entries = map(lambda x: (stat(path(x)).st_size, x), file_list)

        # Sort by mtime
        # [KN] Note that the sort() is guaranteed to be stable (see
        # https://wiki.python.org/moin/HowTo/Sorting#Sort_Stability_and_Complex_Sorts)
        entries.sort(key=lambda x: x[0], reverse=not self.is_asc)

        # return the short names only
        return map(lambda x: x[1], entries)


class ExtensionSorter(Sorter):
    """
    Sorts the list by extensions. Extensions are ordered in alphanumeric way.
    """
    def sort(self, parent_folder, file_list):
        ext = lambda x: os.path.splitext(x)[1]

        # Create pairs (ext, short_name)
        entries = map(lambda x: (ext(x), x), file_list)

        # Sort by extension
        entries.sort(key=lambda x: x[0], reverse=not self.is_asc)

        # return the short names only
        return map(lambda x: x[1], entries)