"""
Copyright (c) 2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
"""
import os
import re
import magic
import sys

from fileman.middleware.roots.parser import PathSplitter


def get_home_dir(username):
    """
    Returns the path to the home directory of the given user.
    :param username: The username
    :return: unicode string
    """
    return unicode(os.path.expanduser('~%s' % username))


def walk_directory(path):
    """
    Generator function which goes through the directory and yields the files and folder names
    :param path: Path to scan.
    """
    queue = [path]

    while queue:
        path = queue.pop()
        contents = os.listdir(path)

        if contents:
            for item in contents:
                fullname = os.path.join(path, item)
                if os.path.isdir(fullname):
                    queue.append(fullname)
                yield fullname
        else:
            yield path


def is_home_request(path):
    """
    :param path:
    :return: Returns True if and only if the requested path is home directory of the current user (~).
    """
    return re.match(r'^/*~/*$', path)


def enumerate_files(rel_paths):
    for name in rel_paths:
        if not os.path.isdir(name):
            yield name
        else:
            for file_name in walk_directory(name):
                yield file_name


def get_filelist(path):
    """
    :param path: Path to scan.
    :return: List of files within the given folder and below it.
    """
    queue = [path]

    result = []
    while queue:
        path = queue.pop()
        contents = os.listdir(path)

        for item in contents:
            fullname = os.path.join(path, item)
            if os.path.isdir(fullname):
                queue.append(fullname)
            else:
                result.append(fullname)

    return result


def rstrip_url(url):
    """
    Removes trailing slashes at the end of the given URL.
    :param url:
    :return:
    """
    return re.sub(r'^(.*?)(/+)$', r'\1', url)


def get_bool(bool_value):
    return 'T' if bool_value else 'F'


def to_bool(value):
    return value == 'T'


def is_property_equal(prop_left, prop_right):
    # [KN] This function is a workaround for the bug in wsgidav: property names are reported with
    # the namespace as the prefix. This function helps to ignore the namespace and compare the property names only:
    # '{DAV:}propname' and '{www.sameschema.org:}propname' will be reported as equal.
    regex = r'^\{.*\}(.*)$'
    l_value = re.sub(regex, r'\1', prop_left)
    r_value = re.sub(regex, r'\1', prop_right)
    return l_value == r_value


MAGIC_MIME_TYPE_WITH_CHARSET_STR_PATTERN = '^(.*); charset=(.*)$'
MAGIC_MIME_TYPE_STR_PATTERN = '^[-\w\+]+/[-\w\+]+$'


def parse_type_and_encoding(raw_str):
    m = re.search(MAGIC_MIME_TYPE_WITH_CHARSET_STR_PATTERN, raw_str)
    if m:
        return m.group(1), m.group(2)

    m = re.match(MAGIC_MIME_TYPE_STR_PATTERN, raw_str)
    if m:
        return raw_str, ''

    return '', ''


def _sanitize_unicode_path(path):
    if isinstance(path, unicode):
        path = path.encode('utf-8')
    return path


def get_mime_type(path):
    ms = magic.open(magic.MAGIC_MIME)
    ms.load()
    normpath = os.path.normpath(path)
    mime_and_encoding = ms.file(_sanitize_unicode_path(normpath))
    mime_type, encoding = parse_type_and_encoding(mime_and_encoding)

    return mime_type


def get_file_encoding(path):
    ms = magic.open(magic.MAGIC_MIME)
    ms.load()
    normpath = os.path.normpath(path)
    mime_and_encoding = ms.file(_sanitize_unicode_path(normpath))
    mime_type, encoding = parse_type_and_encoding(mime_and_encoding)

    return encoding


def is_archive(path):
    ms = magic.open(magic.MAGIC_MIME)
    ms.load()
    normpath = os.path.normpath(path)
    mime = ms.file(_sanitize_unicode_path(normpath))
    return 'application/zip' in mime


def to_relative_path(abs_path, new_root):
    """
    Converts the absolute path to a relative (rooted by new_root). Example
    to_relative_path('/home/hpotter/test/', '/home') will return hpotter/test/
    :param abs_path: The source absolute path
    :param new_root: The new root where the relative must be constructed from.
    :return: The relative path. Note: ValueError is thrown if the root path is not a part of abs_path.
    """
    splitter = PathSplitter()
    abs_fragments = splitter.parse(abs_path)
    root_fragments = splitter.parse(new_root)

    for fragment in root_fragments:
        if abs_fragments[0] != fragment:
            raise ValueError("Impossible to construct the relative path: The root must be the part of the "
                             "absolute path.")
        abs_fragments.pop(0)
    return '/'.join(abs_fragments)
