#!/usr/bin/env bash 

#===============================================================================
#
#          FILE:  ac-cert-tool.sh
#
#   DESCRIPTION:
#                 Auxiliary tool that helps to manage SSL certificates in Adaptive
#                 Computing products
#
#         USAGE:  ./ac-cert-tool.sh <command> [command options]
#
#
#  REQUIREMENTS:  'openssl' util should be installed. Some commands require root
#                 access
#
#         NOTES:  ---
#===============================================================================

set -o nounset                              # Treat unset variables as an error
set -o errexit                              # Exit when any commands fails
set -o pipefail                             # Catch pipe errors

#===============================================================================
#                               Variables
#===============================================================================

PROG=ac-cert-tool.sh

CWD=${0%/*}
TOOLS_DIR="$CWD/tools"
TEMPLATES_DIR="$CWD/templates"

OUT_DIR="$CWD/certs"

CA_DIR="$OUT_DIR/ca"
CA_KEY="$CA_DIR/ca-key.pem"
CA_CSR="$CA_DIR/ca-csr.pem"
CA_CERT="$CA_DIR/ca-cert.pem"
CA_CERT_DIR="$CA_DIR/certs"
CA_SERIAL_FILE="$CA_DIR/serial"
CA_CERTINDEX_FILE="$CA_DIR/certindex.txt"

SERVERS_DIR="$OUT_DIR/servers"
SERVER_KEY_NAME="server-key.pem"
SERVER_CSR_NAME="server-csr.pem"
SERVER_CERT_NAME="server-cert.pem"
SERVER_INSTALL_SCRIPT_NAME="install-server-certs.sh"
SERVER_INSTALL_SCRIPT_TEMPLATE="$TEMPLATES_DIR/$SERVER_INSTALL_SCRIPT_NAME"

CLIENT_DIR="$OUT_DIR/client"
CLIENT_KEY_NAME="client-key.pem"
CLIENT_CSR_NAME="client-csr.pem"
CLIENT_CERT_NAME="client-cert.pem"
CLIENT_KEY="$CLIENT_DIR/$CLIENT_KEY_NAME"
CLIENT_CSR="$CLIENT_DIR/$CLIENT_CSR_NAME"
CLIENT_CERT="$CLIENT_DIR/$CLIENT_CERT_NAME"
CLIENT_INSTALL_SCRIPT_NAME="install-client-certs.sh"
CLIENT_INSTALL_SCRIPT_TEMPLATE="$TEMPLATES_DIR/$CLIENT_INSTALL_SCRIPT_NAME"
CLIENT_INSTALL_SCRIPT="$CLIENT_DIR/$CLIENT_INSTALL_SCRIPT_NAME"

#===============================================================================
#                           Check requirements
#===============================================================================

check_requirements()
{
    command -v openssl >/dev/null 2>&1 || { echo >&2 "Script uses openssl tool to handle certificates but it's not installed.  Aborting."; exit 1; }
}

#===============================================================================
#                           Command line parsing
#===============================================================================

show_help()
{
        cat <<EOF
Usage: ./$PROG <command> [command options]

Valid commands are:

  create-ca          - create Certificate Autorithy
  create-server-cert - create File Manager Service server certificate
  create-client-cert - create client certificate
EOF
exit 1
}

if [ "$#" -lt 1 ]; then
    show_help
fi

cmd=$1
case $cmd in
    "create-ca") check_requirements; source $TOOLS_DIR/create-ca.sh;;
    "create-server-cert") check_requirements; source $TOOLS_DIR/create-server-cert.sh;;
    "create-client-cert") check_requirements; source $TOOLS_DIR/create-client-cert.sh;;
    *          ) echo "Wrong command: $1" >&2; show_help;;
esac
