#!/bin/bash

#===============================================================================
#
#          FILE:  create-ca.sh
#
#   DESCRIPTION:
#                 Create Certificate Authority for usage in in Adaptive
#                 Computing products
#
#         USAGE:  ./ac-cert-tool.sh create-ca [command options]
#
#       OPTIONS:  -d, --days      number of days after which certificate will expire
#                 -k, --keylen    RSA key length
#                 -s, --subject   CA certificate subject in form /type0=value0/type1=value1/type2=...,
#	          -h, --help      show help and exit
#
#
#  REQUIREMENTS:  'openssl' util should be installed. Some commands require root
#                 access
#
#         NOTES:  ---
#===============================================================================

#===============================================================================
#                          Variables and defaults
#===============================================================================

COMMAND="create-ca"

SIGNATURE_HASH="sha256"
START_SERIAL="01"

days=3650 # 10 years
keylen=4096
commonName="Adaptive Computing CA"
subj="/CN=$commonName"

#===============================================================================
#                          Command line parsing
#===============================================================================

show_help()
{
        cat <<EOF
Usage: ./$PROG $COMMAND [options]
Valid options are:
        -d, --days                    number of days after which certificate will expire
        -k, --keylen                  RSA key length
        -s, --subject                 CA certificate subject in form /type0=value0/type1=value1/type2=...,
	-h, --help                    show this text and exit
EOF
exit 1
}

TEMP=`getopt -o h,d:,k:,s: --long help,days:,keylen:,subject -n $PROG -- "$@"`
eval set -- "$TEMP"

while :; do
    case "$1" in
	--) shift; break
	    ;;
        -d|--days) shift ; days="$1"
	    ;;
        -k|--keylen) shift ; keylen="$1"
	    ;;
        -s|--subject) shift ; subj="$1"
	    ;;
        -h|--help) show_help
	    ;;
        *) echo "Unknown option" >&2 ; show_help
    esac
    shift
done

#===============================================================================
#                           Create CA certificate and key
#===============================================================================

mkdir -p "$CA_DIR" || { echo >&2 "Can't create output directory: $CA_DIR. Aborting"; exit 1; }

# Check if key and certificate are already generated
if [ -f "$CA_KEY" ]; then
    echo >&2 "Key file $CA_KEY already exists. Remove old certificate and key if you want recreate them. Aborting";
    exit 1;
fi

if [ -f "$CA_CERT" ]; then
    echo >&2 "Certificate file $CA_KEY already exists. Remove old certificate and key if you want to recreate them. Aborting";
    exit 1;
fi

# Generate key
openssl genrsa -out "$CA_KEY" $keylen || { echo >&2 "Can't generate key file. Aborting"; exit 1; }

# Generate openssl config
openssl_cfg="$(command mktemp)"
cat > $openssl_cfg <<EOF
[req]
distinguished_name = req_distinguished_name
req_extensions = v3_req

[req_distinguished_name]
commonName = $commonName

[ v3_req ]
# Extensions to add to a certificate request
basicConstraints = critical,CA:TRUE
keyUsage = critical, keyCertSign
EOF

# Create CSR
openssl req -new "-$SIGNATURE_HASH" -subj "$subj" -out "$CA_CSR" -key "$CA_KEY" -config $openssl_cfg || \
    { echo >&2 "Can't generate CSR file. Aborting"; rm -f $openssl_cfg; exit 1; }

# Generate server certificate
openssl x509 -req "-$SIGNATURE_HASH" -days $days -in "$CA_CSR" -signkey "$CA_KEY" -out "$CA_CERT" -extensions v3_req -extfile $openssl_cfg || \
    { echo >&2 "Can't generate CA certificate. Aborting"; rm -f $openssl_cfg; exit 1; }
rm -f $openssl_cfg
rm -f $CA_CSR

# Create directories and files required by "openssl ca".
mkdir -p "$CA_CERT_DIR" || { echo >&2 "Can't create CA cert directory: $CA_CERT_DIR. Aborting"; exit 1; }
touch "$CA_CERTINDEX_FILE" || { echo >&2 "Can't create CA cert index file: $CA_CERTINDEX_FILE. Aborting"; exit 1; }
echo "$START_SERIAL" > "$CA_SERIAL_FILE" || { echo >&2 "Can't create CA serial file: $CA_SERIAL_FILE. Aborting"; exit 1; }
