#!/bin/bash

#===============================================================================
#
#          FILE:  create-client-cert.sh
#
#   DESCRIPTION:
#                 Create Certificate Authority for usage in in Adaptive
#                 Computing products
#
#         USAGE:  ./ac-cert-tool.sh create-client-cert [command options] <hostname or ip>
#
#       OPTIONS:  -h, --help                    show help and exit
#
#
#  REQUIREMENTS:  'openssl' util should be installed. Some commands require root
#                 access
#
#         NOTES:  ---
#===============================================================================

#===============================================================================
#                          Variables and defaults
#===============================================================================

COMMAND="create-client-cert"

SIGNATURE_HASH="sha256"

days=3650 # 10 years
keylen=4096
commonName="Adaptive Computing WebDav Client"
subj="/CN=$commonName"
altnames=""

#===============================================================================
#                          Command line parsing
#===============================================================================

show_help()
{
        cat <<EOF
Usage: ./$PROG $COMMAND [options] <hostname or ip>
Valid options are:
        -d, --days          number of days after which certificate will expire
        -k, --keylen        RSA key length
        -s, --subject       client certificate subject in form /type0=value0/type1=value1/type2=...,
	-h, --help          show this text and exit
EOF
exit 1
}

TEMP=`getopt -o h,d:,k:,s: --long help,days:,keylen:,subject:: -n $PROG -- "$@"`
eval set -- "$TEMP"

while :; do
    case "$1" in
	--) shift; break
	    ;;
        -d|--days) shift ; days="$1"
	    ;;
        -k|--keylen) shift ; keylen="$1"
	    ;;
        -s|--subject) shift ; subj="$1"
	    ;;
        -h|--help) show_help
	    ;;
        *) echo "Unknown option" >&2 ; show_help
    esac
    shift
done

#===============================================================================
#                       Create client certificate and key
#===============================================================================

mkdir -p "$CLIENT_DIR" || { echo >&2 "Can't create output directory: $CLIENT_DIR. Aborting"; exit 1; }

# Check if key and certificate are already generated
if [ -f "$CLIENT_KEY" ] ; then
    echo >&2 "Key file $CLIENT_KEY already exists. Remove old certificate and key if you want recreate them. Aborting";
    exit 1;
fi

if [ -f "$CLIENT_CERT" ]; then
    echo >&2 "Certificate file $CLIENT_CERT already exists. Remove old certificate and key if you want to recreate them. Aborting";
    exit 1;
fi

# Generate key
openssl genrsa -out "$CLIENT_KEY" $keylen || { echo >&2 "Can't generate key file. Aborting"; exit 1; }

# Generate openssl config
openssl_cfg="$(command mktemp)"
cat > $openssl_cfg <<EOF
[ ca ]
default_ca				= CA_default

[ CA_default ]
dir                                     = $CA_DIR
serial					= $CA_DIR/serial
database				= $CA_DIR/certindex.txt
new_certs_dir				= $CA_DIR/certs
certificate				= $CA_CERT
private_key				= $CA_KEY
default_days				= $days
default_md				= sha256
preserve				= no
email_in_dn				= no
nameopt					= default_ca
certopt					= default_ca
policy					= policy_match
unique_subject                          = no

x509_extensions	= usr_cert		# The extentions to add to the cert

# For the CA policy
[ policy_match ]
countryName		= optional
stateOrProvinceName	= optional
organizationName	= optional
organizationalUnitName	= optional
commonName		= supplied
emailAddress		= optional

[ usr_cert ]

# These extensions are added when 'ca' signs a request.

# This goes against PKIX guidelines but some CAs do it and some software
# requires this to avoid interpreting an end user certificate as a CA.
basicConstraints = critical,CA:FALSE

keyUsage = critical, digitalSignature
extendedKeyUsage=clientAuth

# Client type certificate
nsCertType			= client
EOF

# Create CSR
# Create CSR
openssl req -new -subj "$subj" -out "$CLIENT_CSR" -key "$CLIENT_KEY" || \
    { echo >&2 "Can't generate CSR file. Aborting"; rm -f $openssl_cfg; exit 1; }

# Generate client certificate
openssl ca -batch -config $openssl_cfg -in "$CLIENT_CSR" -out "$CLIENT_CERT" -notext || \
    { echo >&2 "Can't generate client certificate. Aborting"; rm -f $openssl_cfg; exit 1; }
rm -f $openssl_cfg
rm -f $CLIENT_CSR

# Generate install script
client_key="$(command cat $CLIENT_KEY)"
client_cert="$(command cat $CLIENT_CERT)"
ca_cert="$(command cat $CA_CERT)"

cat > $CLIENT_INSTALL_SCRIPT <<EOF
#!/bin/bash
#===============================================================================
#
#          FILE:  $CLIENT_INSTALL_SCRIPT_NAME
#
#   DESCRIPTION:  Install client certificate, key and CA certificate
#
#         USAGE:  Usage: ./$CLIENT_INSTALL_SCRIPT_NAME [options] <target dir>
#
#       OPTIONS:  -k, --clientkey      name of client key file in target directory
#                 -c, --clientcert     name of client cert file in target directory
#                 -a, --cacert         name of CA cert file in target directory
#	          -p, --permissions    permissions for installed files
#                 -u, --user           specify user and group (in form user:group) for installed files
#	          -h, --help           show help and exit
#
#  REQUIREMENTS:  You should have enough permissions to change user:group and set
#                 right permissions for installed files
#
#
#
#===============================================================================

#===============================================================================
#                       Autogenerated Variables
#===============================================================================
client_key="$client_key"

client_cert="$client_cert"

ca_cert="$ca_cert"
EOF

cat $CLIENT_INSTALL_SCRIPT_TEMPLATE >> $CLIENT_INSTALL_SCRIPT
