#!/bin/bash

#===============================================================================
#
#          FILE:  create-server-cert.sh
#
#   DESCRIPTION:
#                 Create Certificate Authority for usage in in Adaptive
#                 Computing products
#
#         USAGE:  ./ac-cert-tool.sh create-server [command options] <hostname or ip>
#
#       OPTIONS:  -d, --days          number of days after which certificate will expire
#                 -k, --keylen        RSA key length
#                 -s, --subject       server certificate subject in form /type0=value0/type1=value1/type2=...,
#                 -a, --altnames      comma separated list of alternative DNS names.
#	          -h, --help          show this text and exit
#
#  REQUIREMENTS:  'openssl' util should be installed. Some commands require root
#                 access
#
#         NOTES:  ---
#===============================================================================

#===============================================================================
#                          Variables and defaults
#===============================================================================

COMMAND="create-server-cert"

SIGNATURE_HASH="sha256"

days=3650 # 10 years
keylen=4096
subj="/CN=Adaptive Computing File Manager Service"
altnames=""

#===============================================================================
#                          Auxiliary functions
#===============================================================================

valid_ip()
{
    local  ip=$1
    local  stat=1

    if [[ $ip =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$ ]]; then
        OIFS=$IFS
        IFS='.'
        ip=($ip)
        IFS=$OIFS
        [[ ${ip[0]} -le 255 && ${ip[1]} -le 255 \
            && ${ip[2]} -le 255 && ${ip[3]} -le 255 ]]
        stat=$?
    fi
    return $stat
}

#===============================================================================
#                          Command line parsing
#===============================================================================

show_help()
{
        cat <<EOF
Usage: ./$PROG $COMMAND [options] <hostname or ip>
Valid options are:
        -d, --days          number of days after which certificate will expire
        -k, --keylen        RSA key length
        -s, --subject       server certificate subject in form /type0=value0/type1=value1/type2=...,
        -a, --altnames      comma separated list of alternative DNS names.
	-h, --help          show this text and exit
EOF
exit 1
}

TEMP=`getopt -o h,d:,k:,s:,a: --long help,days:,keylen:,subject:,altnames: -n $PROG -- "$@"`
eval set -- "$TEMP"

while :; do
    case "$1" in
	--) shift; break
	    ;;
        -d|--days) shift ; days="$1"
	    ;;
        -k|--keylen) shift ; keylen="$1"
	    ;;
        -s|--subject) shift ; subj="$1"
	    ;;
        -a|--altnames) shift ; altnames="$1"
	    ;;
        -h|--help) show_help
	    ;;
        *) echo "Unknown option" >&2 ; show_help
    esac
    shift
done

if [ "$#" -lt 2 ]; then
    echo >&2 "<hostname> argument is required";
    show_help
fi

hostname=$2
out_dir="$SERVERS_DIR/$hostname"
key_file="$out_dir/$SERVER_KEY_NAME"
csr_file="$out_dir/$SERVER_CSR_NAME"
cert_file="$out_dir/$SERVER_CERT_NAME"
install_script="$out_dir/$SERVER_INSTALL_SCRIPT_NAME"

#===============================================================================
#                       Create server certificate and key
#===============================================================================

mkdir -p "$out_dir" || { echo >&2 "Can't create output directory: $out_dir. Aborting"; exit 1; }

# Check if key and certificate are already generated
if [ -f "$key_file" ] ; then
    echo >&2 "Key file $key_file already exists. Remove old certificate and key if you want recreate them. Aborting";
    exit 1;
fi

if [ -f "$cert_file" ]; then
    echo >&2 "Certificate file $cert_file already exists. Remove old certificate and key if you want to recreate them. Aborting";
    exit 1;
fi

# Generate key
openssl genrsa -out "$key_file" $keylen || { echo >&2 "Can't generate key file. Aborting"; exit 1; }

# Generate openssl config
openssl_cfg="$(command mktemp)"
cat > $openssl_cfg <<EOF
[ ca ]
default_ca				= CA_default

[ CA_default ]
dir                                     = $CA_DIR
serial					= $CA_DIR/serial
database				= $CA_DIR/certindex.txt
new_certs_dir				= $CA_DIR/certs
certificate				= $CA_CERT
private_key				= $CA_KEY
default_days				= $days
default_md				= sha256
preserve				= no
email_in_dn				= no
nameopt					= default_ca
certopt					= default_ca
policy					= policy_match
unique_subject                          = no

x509_extensions	= usr_cert		# The extentions to add to the cert

# For the CA policy
[ policy_match ]
countryName		= optional
stateOrProvinceName	= optional
organizationName	= optional
organizationalUnitName	= optional
commonName		= supplied
emailAddress		= optional

[ usr_cert ]

# These extensions are added when 'ca' signs a request.

# This goes against PKIX guidelines but some CAs do it and some software
# requires this to avoid interpreting an end user certificate as a CA.
basicConstraints = critical,CA:FALSE

keyUsage = critical, digitalSignature, keyEncipherment
extendedKeyUsage=serverAuth,clientAuth

# Server type certificate
nsCertType			= server

# Set alternative names
subjectAltName = @alt_names

[alt_names]
EOF

ip_counter=1
dns_counter=1

add_altname()
{
    local altname=$1
    if valid_ip $altname; then
	#=======================================================================
	# HACK!!! Unfortunately some python implementations of SSL hostname verification
	# (e.g. match_hostname() function in 'requests' library) don't fully implement
	# RFC 2818 and RFC 6125 and ignore IPs in SSL certificate. So we will put them
	# as usual domains.
	#=======================================================================
	echo "DNS.$dns_counter = $altname" >> $openssl_cfg
	(( dns_counter++ ))
	#=======================================================================
	echo "IP.$ip_counter = $altname" >> $openssl_cfg
	(( ip_counter++ ))
    else
	echo "DNS.$dns_counter = $altname" >> $openssl_cfg
	(( dns_counter++ ))
    fi
}

add_altname $hostname

if [ -n "$altnames" ]; then
    IFS=',' read -ra ALT_NAMES <<< "$altnames"
    for altname in "${ALT_NAMES[@]}"; do
	add_altname $altname
    done
fi

# Create CSR
openssl req -new -subj "$subj" -out "$csr_file" -key "$key_file" || \
    { echo >&2 "Can't generate CSR file. Aborting"; rm -f $openssl_cfg; exit 1; }

# Generate server certificate
openssl ca -batch -config $openssl_cfg -in "$csr_file" -out "$cert_file" -notext || \
    { echo >&2 "Can't generate server certificate. Aborting"; rm -f $openssl_cfg; exit 1; }
rm -f $openssl_cfg
rm -f $csr_file

# Generate install script
server_key="$(command cat $key_file)"
server_cert="$(command cat $cert_file)"
ca_cert="$(command cat $CA_CERT)"

cat > $install_script <<EOF
#!/bin/bash
#===============================================================================
#
#          FILE:  $SERVER_INSTALL_SCRIPT_NAME
#
#   DESCRIPTION:  Install server certificate, key and CA certificate
#
#         USAGE:  Usage: ./$SERVER_INSTALL_SCRIPT_NAME [options] <target dir>
#
#       OPTIONS:  -k, --serverkey      name of server key file in target directory
#                 -c, --servercert     name of server cert file in target directory
#                 -a, --cacert         name of CA cert file in target directory
#	          -p, --permissions    permissions for installed files
#                 -u, --user           specify user and group (in form user:group) for installed files
#	          -h, --help           show help and exit
#
#  REQUIREMENTS:  You should have enough permissions to change user:group and set
#                 right permissions for installed files
#
#
#===============================================================================

#===============================================================================
#                       Autogenerated Variables
#===============================================================================
server_key="$server_key"

server_cert="$server_cert"

ca_cert="$ca_cert"
EOF

cat $SERVER_INSTALL_SCRIPT_TEMPLATE >> $install_script
